"""Custom Types."""

from datetime import datetime, timedelta
from typing import Literal, Union

JSON_LD = dict
ISO8601_Date = str
Date = Union[ISO8601_Date, datetime]
URL = str
ISO8601_Duration = str  # ISO8601Time https://en.wikipedia.org/wiki/ISO_8601#Durations
Duration = Union[ISO8601_Duration, timedelta]
InstructionType = Union[Literal["HowToSection"], Literal["HowToStep"]]



"""RecipeSage data models."""

import re
from dataclasses import dataclass
from typing import Iterable, Optional, Type

from recipe_scrapers import AbstractScraper

from ._types import JSON_LD, URL, Date, Duration, InstructionType
from .utils import simple_ISO8601_duration


@dataclass
class Instruction:
    """Recipe instruction."""

    _type: InstructionType
    text: str

    def to_json_ld(self: "Instruction") -> JSON_LD:
        """Export to JSON-LD.

        Returns:
            JSON_LD: output
        """
        return {"@type": self._type, "text": self.text}


@dataclass
class Comment:
    """Recipe comment."""

    name: str
    text: str

    def to_json_ld(self: "Comment") -> JSON_LD:
        """Export to JSON-LD.

        Returns:
            JSON_LD: output
        """
        return {
            "@type": "Comment",
            "name": self.name or "Author Notes",
            "text": self.text,
        }


@dataclass
class RecipeSage:
    """RecipeSage data model."""

    datePublished: Optional[Date]
    description: Optional[str]
    image: Iterable[URL]
    name: str
    prepTime: Optional[Duration]
    recipeIngredient: Iterable[str]
    recipeInstructions: Iterable[Instruction]
    recipeYield: Optional[str]
    totalTime: Optional[Duration]
    recipeCategory: Iterable[str]
    creditText: Optional[str]
    isBasedOn: Optional[URL]
    comment: Iterable[Comment]

    @classmethod
    def from_scraper(cls: "Type[RecipeSage]", scraper: AbstractScraper) -> "RecipeSage":
        """Create RecipeSage from scraper.

        Args:
            scraper (AbstractScraper): scraper to convert

        Returns:
            RecipeSage: RecipeSage model
        """

        _instructions = []

        _ins = scraper.schema.data.get("recipeInstructions", [])

        if isinstance(_ins, str):
            _ins = re.sub(r"\n+", "\n", _ins).splitlines()

        for ins in _ins:
            if isinstance(ins, dict):
                _instructions.append(
                    Instruction(ins.get("@type", "HowToStep"), ins.get("text", ""))
                )
            elif isinstance(ins, str):
                _instructions.append(Instruction("HowToStep", ins))
            else:
                message = f"got a bad value: {ins}"
                raise ValueError(message)

        return cls(
            datePublished=scraper.schema.data.get("datePublished", None),
            description=scraper.schema.data.get("description", None),
            image=scraper.schema.image(),
            name=scraper.schema.data.get("name"),
            prepTime=scraper.schema.data.get("prepTime", None),
            recipeIngredient=scraper.schema.ingredients(),
            recipeInstructions=_instructions,
            recipeYield=scraper.schema.data.get("recipeYield", None),
            totalTime=scraper.schema.data.get("totalTime", None),
            recipeCategory=scraper.schema.data.get("recipeCategory", []),
            creditText=None,
            isBasedOn=scraper.url,
            comment=[],
        )

    def to_json_ld(self: "RecipeSage") -> JSON_LD:
        """Convert to JSON-LD for import/export.

        Returns:
            JSON_LD: JSON-LD data model
        """

        _prepTime: str = simple_ISO8601_duration(self.prepTime) if self.prepTime else ""
        _totalTime: str = (
            simple_ISO8601_duration(self.totalTime) if self.totalTime else ""
        )

        return {
            "@context": "http://schema.org",
            "@type": "Recipe",
            "datePublished": self.datePublished or "",
            "description": self.description or "",
            "image": self.image,
            "name": self.name,
            "prepTime": _prepTime,
            "recipeIngredient": [
                ingredient for ingredient in self.recipeIngredient if ingredient
            ],
            "recipeInstructions": [
                instruction.to_json_ld() for instruction in self.recipeInstructions
            ],
            "recipeYield": self.recipeYield or "",
            "totalTime": _totalTime,
            "recipeCategory": self.recipeCategory or [],
            "creditText": self.creditText or "",
            "isBasedOn": self.isBasedOn or "",
            "comment": [comment.to_json_ld() for comment in self.comment],
        }



"""Convert recipe scrapers to RecipeSage models appropriate for import/export."""

from recipe_scrapers import AbstractScraper

from ._types import JSON_LD
from .models import RecipeSage


def export_recipe(scraper: AbstractScraper) -> JSON_LD:
    """Export recipe to RecipeSage JSON-LD format.

    Args:
        scraper (AbstractScraper): recipe scraper

    Returns:
        JSON_LD: output
    """
    return RecipeSage.from_scraper(scraper).to_json_ld()


__all__ = ["export_recipe", "RecipeSage"]



"""Utilities."""

import re
from datetime import timedelta

from ._types import Duration, ISO8601_Duration

ISO8601_DURATION_REGEX = re.compile(
    r"^P(?!$)"
    r"(?P<PY>\d+(?:\.\d+)?Y)?"
    r"(?P<PM>\d+(?:\.\d+)?M)?"
    r"(?P<PW>\d+(?:\.\d+)?W)?"
    r"(?P<PD>\d+(?:\.\d+)?D)?"
    r"(T(?=\d)"
    r"(?P<TH>\d+(?:\.\d+)?H)?"
    r"(?P<TM>\d+(?:\.\d+)?M)?"
    r"(?P<TS>\d+(?:\.\d+)?S)?"
    r")?$"
)


def simple_ISO8601_duration(duration: Duration) -> ISO8601_Duration:
    """Create a simple ISO8601 Duration importable by RecipeSage.

    Notes: RecipeSage only allows the `PT<N.NH><N.NM><N.NS>` format.
        See [here](https://github.com/julianpoy/RecipeSage/blob/679a177f19d2f7be2aadaab50c3e8c3b4c4d7b94/packages/backend/src/services/json-ld.js#L151-L157).

    Args:
        duration (Duration): duration to format

    Returns:
        ISO8601_Duration: simple duration
    """

    def seconds_to_ISO8601_duration(_time: int) -> ISO8601_Duration:
        out = "PT"
        seconds = _time % 60
        _time = _time // 60
        minutes = _time % 60
        hours = _time // 60
        if hours > 0:
            out += f"{hours}H"
        if minutes > 0:
            out += f"{minutes}M"
        if seconds > 0:
            out += f"{seconds}S"
        return out

    if isinstance(duration, timedelta):
        return seconds_to_ISO8601_duration(int(duration.total_seconds()))

    if isinstance(duration, str):
        _match = ISO8601_DURATION_REGEX.match(duration)
        if _match:
            _time = 0.0
            _time += float((_match.groupdict().get("TS") or "0").strip("S"))
            _time += float((_match.groupdict().get("TM") or "0").strip("M")) * 60
            _time += float((_match.groupdict().get("TH") or "0").strip("H")) * 3600
            _time = int(_time)

            return seconds_to_ISO8601_duration(_time)
        message = f"not a valid ISO8601_Duration: {duration}"
        raise ValueError(message)
    message = f"not a valid Duration: {duration}"
    raise ValueError(message)
